#!/bin/python3

import sys

if len(sys.argv) <2:
  print( "usage:" )
  print("  " + sys.argv[0] + "GELname", "[Folder-prefix=$Device-$Version-$UpdateType]" )
  print( "  " )
  exit()

import os
from binascii import crc32
from struct import unpack

try:
  from PIL import Image
except:
  Image=None

#AFAIK only Python > 3.3 has the proper decompressor...
# also something like pyliblzma is probably needed
try:
  import lzma
  decompressor = lzma.LZMADecompressor(lzma.FORMAT_ALONE)  ##test functionality
except:
  decompressor=None

def main():
  global prefix, decompressor
  #get unpack file
  src_file_path = sys.argv[1]

  #read file length
  src_file_size = os.path.getsize(src_file_path)
  #read file
  b = bytearray(open(src_file_path, 'rb').read())

  ## Start decoding

  instrument = toStr(b[0:16])
  print("instrument series:\t"+ instrument)
  version = toStr(b[16:32])
  print("firmware version:\t"+ version)
  pos = 32

  updateType = b[pos:pos+4] # ???
  pos +=4
  print ("updateType:\t0x" + "".join(["%02x"%i for i in updateType]))

  #prefix = sys.argv[2] if len(sys.argv)>=3 else "."
  prefix = instrument + "-" + version + "-" + str(updateType[1])


  nHeaders = word2int(b[pos:pos+4]) # 
  pos +=4

  # crude way to parse the directory table
  files = []
  print ("found %i files\n"%nHeaders)
  for i in range(nHeaders):
    h = getHeader(b[pos:pos+60])
    pos +=60
    files.append(h)
    print(h)

  #write entire header
  save("/header", b[:pos])


  for h in files:
    bFile = b[h["start"]:h["start"]+h["length"]]

    #CRC check
    crc = crc32(bFile)& 0xffffffff #crude workaround because its interpreted as unsigned
    if crc == h["crc"]:   # write Files

      #read Header
      fileCRC = word2int(bFile[0:4])
      fileType = word2int(bFile[4:8])
      fileLength = word2int(bFile[8:12])
      p1 = word2int(bFile[12:16])
      firmware = word2int(bFile[16:20])

      # crude sanity check
      if fileLength== h["length"]-24:
        crc2 = crc32(bFile[24:]) & 0xffffffff
        if fileCRC == crc2:
          #write header file
          save(h["filename"] + ".header",bFile[:24])
          
          #write stripped file
          save(h["filename"],bFile[24:])
          #write decompressed file
          if decompressor != None:
            if bFile[24:28] == bytearray([0x5D, 0,0, 0x80]): #LZMA Magic value
              decompressor = lzma.LZMADecompressor(lzma.FORMAT_ALONE) #restart decompressor
              try:
                dec = decompressor.decompress(bFile[24:])
                if len(dec)>0:
                  save(h["filename"]+".decompressed",dec)
              except lzma.LZMAError as e:
                print(e)
                pass
        else:
          print ("secondary CRC mismatch: '%s' (0x%08x vs 0x%08x)"%(h["filename"], fileCRC, crc2))

      else:
        #write Raw file
        save(h["filename"],bFile)

    else: # don't write the file if CRC mismatch
      print ("CRC mismatch: '%s' (0x%08x vs 0x%08x)"%(h["filename"], h["crc"], crc))
      if h["filename"] == "":
        save("/footer",b[h["start"]:h["start"]+h["length"]])
        if  Image:
          img = b[h["start"]+5*4:h["start"]+h["length"]-4]
          img.reverse()
          Image.frombuffer("L", (8, int((len(img)/8))), img, "raw", "L", 0,1).save(prefix + "/footer.png")
            
  #  print(pos, h["start"], h["start"]-pos)  # check if bytes are missed
    pos += h["length"]

  print("")
  print ("original filesize:\t%d"%src_file_size)
  print ("bytes processed:\t%d"%pos)


# crude bytearray to String converter
def toStr(b):
  return "".join(["%c"%i for i in b]).strip("\x00")

# convert 4 bytes to int
def word2int(w):
  return unpack("<I", w)[0]

# crude filetable parser
def getHeader(b):
  filename = toStr(b[0:32])
  length = word2int(b[32:36])

  start = word2int(b[36:40])
  checksum = word2int(b[40:44])
  p1 = word2int(b[44:48]) #??? ==0
  p2 = word2int(b[48:52]) #??? ==0
  filetype = word2int(b[52:56]) #???
  p4 = word2int(b[56:60]) #??? ==0
  return {
    "filename": filename,
    "type" : filetype,
    "start" : start,
    "length" : length,
    "crc" : checksum,
    "unknown" : (p1,p2,p4)
    }

#crude way to create nonexisting directories
def checkCreateDir(filename):
  if not os.path.exists(os.path.dirname(filename)):
    try:
        os.makedirs(os.path.dirname(filename))
    except OSError as exc: # Guard against race condition
        if exc.errno != errno.EEXIST:
            raise
            
def save(filename, data):
  checkCreateDir(prefix + filename) #create missing directories
  print ("writing %s\t (%d bytes)"%(filename,len(data)))
  open(prefix + filename, 'wb').write(data)



if __name__=="__main__":
   main()
